package HTTP::Server::Directory;
use warnings;
use strict;

use Log::Report 'httpd-multiplex', syntax => 'SHORT';

use Net::CIDR    qw/cidrlookup/;
use File::Spec   ();

sub _allow_cleanup($);
sub _allow_match($$$$);
sub _filename_trans($$);

=chapter NAME
HTTP::Server::Directory - describe a server directory 

=chapter SYNOPSIS
 # implicit creation of ::Directory object
 my $vh = HTTP::Server::VirtualHost
   ->new(directories => {path => '/', location => ...})

 # explicit use
 my $root = HTTP::Server::Directory
   ->new(path => '/', location => '...');
 my $vh = HTTP::Server::VirtualHost
   ->new(directories => $root);

=chapter DESCRIPTION
Each M<HTTP::Server::VirtualHost> will define where the files are
located.  Parts of the URI path can map on different directories,
with different permissions.

User directories, like used in the URI C<<http://xx/~user/yy>>
is implemented in M<HTTP::Server::Directory::UserDirs>.

=chapter METHODS

=section Constructors

=c_method new OPTIONS|HASH-of-OPTIONS

=option   path STRING
=default  path '/'

=requires location DIRECTORY|CODE
The absolute prefix DIRECTORY befor the path of the URI, or a CODE
reference which will rewrite the path (only parameter) into the
absolute file or directory name.

=option   allow   CIDR|HOSTNAME|DOMAIN|CODE|ARRAY
=default  allow   <undef>
Allow all requests which pass any of these parameters, and none
of the deny parameters.  See L</Allow access>.

=option   deny    CIDR|HOSTNAME|DOMAIN|CODE|ARRAY
=default  deny    <undef>
See C<allow> and L</Allow access>

=cut

sub new(@)
{   my $class = shift;
    my $args  = @_==1 ? shift : {@_};
    (bless {}, $class)->init($args);
}

sub init($)
{   my ($self, $args) = @_;

    my $path = $self->{HSD_path}  = $args->{path} || '/';
    my $loc  = $args->{location}
        or error __x"directory definition requires location";

    if(ref $loc eq 'CODE') {;}
    else
    {   File::Spec->file_name_is_absolute($loc)
           or error __x"directory location {loc} for path {path} not absolute"
                , loc => $loc, path => $path;
        -d $loc
           or error __x"directory location {loc} for path {path} does not exist"
                , loc => $loc, path => $path;

        substr($loc,-1) eq '/' or $loc .= '/';
    }
    $self->{HSD_loc}   = $loc;
    $self->{HSD_fn}    = _filename_trans $path, $loc;

    $self->{HSD_allow} = _allow_cleanup $args->{allow};
    $self->{HSD_deny}  = _allow_cleanup $args->{deny};

    $self;
}

#-----------------
=section Attributes
=method path
=method location
=cut

sub path()     {shift->{HSD_path}}
sub location() {shift->{HSD_location}}

#-----------------
=section Permissions

=method allow CLIENT, SESSION, REQUEST, URI
BE WARNED that the URI is the rewrite of the REQUEST uri, and therefore
you should use that URI.  The SESSION represents a user.

See L</Allow access>.
=cut

sub allow($$$$)
{   my ($self, $client, $session, $req, $uri) = @_;
    if(my $allow = $self->{HSD_allow})
    {   $self->_allow_match($client, $session, $uri, $allow) or return 0;
    }
    if(my $deny = $self->{HSD_deny})
    {    $self->_allow_match($client, $session, $uri, $deny) and return 0;
    }
    1;
}

sub _allow_match($$$$)
{   my ($self, $client, $session, $uri, $rules) = @_;
    my ($ip, $host) = @$client{'ip', 'host'};
    first { $_->($ip, $host, $session, $uri) } @$rules ? 1 : 0;
}

sub _allow_cleanup($)
{   my $p = shift or return;
    my @p;
    foreach my $r (ref $p eq 'ARRAY' ? @$p : $p)
    {   push @p
          , ref $r eq 'CODE'    ? $r
          : index($r, ':') >= 0 ? sub {cidrlookup $_[0], $r}    # IPv6
          : $r !~ m/[a-zA-Z]/   ? sub {cidrlookup $_[0], $r}    # IPv4
          : $r =~ s/^\.//       ? sub {$_[1] =~ qr/(^|\.)\Q$r\E$/i} # Domain
          :                       sub {lc($_[1]) eq lc($r)}     # hostname
    }
    @p ? \@p : undef;
}

=method filename PATH
Convert a URI PATH into a directory path.  Return C<undef> if not possible.
=cut

sub filename($) { $_[0]->{HSD_fn}->($_[1]) }

sub _filename_trans($$)
{   my ($path, $loc) = @_;
    return $loc if ref $loc eq 'CODE';
    sub
      { my $x = shift;
        $x =~ s!^\Q$path!$loc! or panic "path $x not within $path";
        $x;
      };
}

=chapter DETAILS

=section Directory limits

=subsection Allow access

The M<allow()> method handles access rights.  When a trueth value is
produced, then access is permitted.

The base class implements access rules via the C<allow> or C<deny>
option of M<new()>.  These parameters are exclusive (which is slightly
different from Apache); you can either allow or deny, but not both at
the same time.

The parameters to C<allow> or C<deny> are an ARRAY with any combination of
=over 4
=item IPv4 and IPv6 addresses
=item IPv4 and IPv6 address ranges in CIDR notation
=item hostname
=item domain name (leading dot)
=item your own CODE reference, which will be called with the IP address,
  the hostname, the session, and the rewritten URI.
=back

=example new(allow) parameters
 MyVHOST->new( allow =>
    [ '192.168.2.1'           # IPv4
    , '10/32'                 # IPv4 CIDR
    , '10.0.0.0-10.3.255.255' # IPv4 range
    , '::dead:beef:0:0/110'   # IPv6 range
    , 'www.example.com'       # hostname
    , '.example.com'          # domain and subdomains
    , 'example.com'           # only this domain
    ], ...

=example create own access rules
If you have an ::VirtualHost extension class, you do this:

 sub allow($$$)
 {   my ($self, $session, $request, $uri) = @_;

     # General rules may refuse access already
     $self->SUPER::allow($session, $request, $uri)
         or return 0;

     # here your own checks
     # $session is a HTTP::Server::Session
     # $request is a HTTP::Request
     # $uri     is a URI::

     1;
 }

You may also pass a code-ref to M<new(allow)>:

 HTTP::Server::VirtualHost->new(allow => \&my_rules);

 sub my_rules($$$$)   # called before each request
 {   my ($ip, $host, $session, $uri) = @_;
     # return true if access is permitted
 }
=cut

1;
