package HTTP::Server::VirtualHost;
use warnings;
use strict;

use HTTP::Server::Multiplex;
use HTTP::Server::Directory;
use HTTP::Server::Directory::UserDirs;

use Log::Report 'httpd-multiplex', syntax => 'SHORT';

use HTTP::Status;
use List::Util   qw/first/;
use English      qw/$EUID/;
use File::Spec   ();

=chapter NAME
HTTP::Server::VirtualHost - base-class for virtual host definitions

=chapter SYNOPSIS
 use HTTP::Server::Multiplex;

 my $vhost  = HTTP::Server::VirtualHost->new(@vhost_opts);
 my $daemon = HTTP::Server::Multiplex->new
   ( @other_options
   , vhosts => $vhost
   );

 # or
 my $daemon = HTTP::Server::Multiplex->new(@other_opts);
 $daemon->addVirtualHost($vhost);
 $daemon->addVirtualHost(@vhost2_opts);

 # create object which extends HTTP::Server::VirtualHost
 my $myvhost = MyVHost->new(...);
 $daemon->addVirtualHost($myvhost);

=chapter DESCRIPTION
These virtual host definitions are used by M<HTTP::Server::Multiplex>, to
implement (server) name based data seperation.  Its features resemble those
of Apache virtual hosts.

Each virtual host usually has to M<HTTP::Server::Directory> slaves: one
which describes the permissions for user directories (url paths in the
form C< /~user/ >) and one for data outside the user space.

=chapter METHODS

=section Constructors
You may avoid the creation of extension classes for each virtual host,
by using these options.

=c_method new OPTIONS|HASH-of-OPTIONS
=requires name    HOSTNAME

=option   aliases HOSTNAME|ARRAY-of-HOSTNAMES
=default  aliases []

=option   rewrite CODE
=default  rewrite <undef>
See M<rewrite()>.

=option   documents DIRECTORY
=default  documents <undef>
An absolute DIRECTORY for the location of the source files.  Creates the
most free M<HTTP::Server::Directory> object.  If you need things like
access restrictions, then do not use this option but the C<directories>
option.

=option   directories OBJECT|HASH|ARRAY
=default  directories <see text>
Pass one or more M<HTTP::Server::Directory> OBJECTS, or HASHes which will
be used to initialize them.  If no information is provided, then only C<<
/~user >> urls and scripts can be used.

=option   user_dirs undef|OBJECT|HASH
=default  user_dirs <see text>
If not specified, a M<HTTP::Server::Directory::UserDirs> is created for
you, with standard Apache behavior.  You may provide your own OBJECT or
a HASH which contains the parameters to create it.  When you explicitly
specify an C<undef> value, then user directories will not be allowed.

=option   index_file STRING|ARRAY
=default  index_file ['index.html', 'index.htm']
When a directory is addressed, it is scanned whether one of these files
exist.  If so, the content will be shown.

=option   directory_list BOOLEAN
=default  directory_list <false>
Enables the display of a directory, when it does not contain one of the
C<index_file> prepared defaults.

=option   handlers HASH
=default  handlers {}
The keys are path names, part of the request URIs.  The values are
CODE-references, called when that URI is addressed.  The access rules
are taken from the directory definition which is selected by the path,
for that's all.

The handlers are called with a the connection (M<HTTP::Server::Connection>),
request (M<HTTP::Request>), uri object (M<URI>).
=cut

sub new(@)
{   my $class = shift;
    my $args  = @_==1 ? shift : {@_};
    (bless {}, $class)->init($args);
}

sub init($)
{   my ($self, $args) = @_;

    my $name = $self->{HSV_name} = $args->{name};
    defined $name
        or error __x"virtual host {pkg} has no name", pkg => ref $self;

    my $aliases = $args->{aliases}            || [];
    $self->{HSV_aliases}  = ref $aliases eq 'ARRAY' ? $aliases : [$aliases];

    $self->{HSV_rewrite}  = $args->{rewrite}  || sub {$_[0]};
    $self->{HSV_dirlist}  = $args->{directory_list};
    $self->{HSV_handlers} = $args->{handlers} || {};

    $self->{HSV_dirs}     = {};
    if(my $docroot = $args->{documents})
    {   File::Spec->file_name_is_absolute($docroot)
            or error __x"vhost {name} documents directory must be absolute"
                 , name => $name;
        -d $docroot
            or error __x"vhost {name} documents `{dir}' must point to dir"
                 , name => $name, dir => $docroot;
        $docroot =~ s/\\$//; # strip trailing / if present
        $self->addDirectory(path => '/', location => $docroot);
    }
    my $dirs = $args->{directories} || [];
    $self->addDirectory($_) for ref $dirs eq 'ARRAY' ? @$dirs : $dirs;

    my $ud;
    if(!exists $args->{user_dirs})
    {   $ud = HTTP::Server::Directory::UserDirs->new }
    elsif($ud = $args->{user_dirs})
    {   if(ref $ud eq 'HASH')
        {   $ud = HTTP::Server::Directory::UserDirs->new($ud) }
        elsif(not $ud->isa('HTTP::Server::Directory::UserDirs'))
        {   error __x"vhost {name} user_dirs is not an ::UserDirs object"
              , name => $self->name;
        }
    }
    $self->{HSV_udirs} = $ud;

    my $if = $args->{index_file};
    my @if = ref $if eq 'ARRAY' ? @$if
           : defined $if        ? $if
           : qw/index.html index.html/;
    $self->{HSV_indexfns} = \@if;

    $self;
}

#---------------------
=section Attributes

=method name
Returns the primary name for this server.

=method aliases
Returns a list of all aliases (alternative names) for this server.
=cut

sub name()    {shift->{HSV_name}}
sub aliases() {@{shift->{HSV_aliases}}}

#---------------------
=section Handler

=method requestForMe URI
Re-check whether a request is really for this virtual host.
=cut

sub requestForMe($)
{   my ($self, $uri) = @_;
    my $host = $uri->host;
    $host eq $self->name || first {$host eq $_} $self->aliases;
}

=method handleRequest CONNECTION, REQUEST
=cut

sub handleRequest($$)
{   my ($self, $conn, $req) = @_;

    my $uri = $self->rewrite($req->uri);
    if($uri ne $req->uri)
    {   info $req->id." rewritten to $uri";
        $self->requestForMe($uri)
            or return $conn->sendRedirect($req, RC_TEMPORARY_REDIRECT, $uri);
    }

    my $path = $uri->path;
    my $tree = $self->directoryOf($path)
        or return $conn->sendStatus($req, RC_FORBIDDEN, "$path not configured");

    $tree->allow($conn->client, $conn->session, $req, $uri)
        or return
             $conn->sendStatus($req, RC_FORBIDDEN, "$path access not allowed");

    if(my $handler = $self->{HSV_handlers}{$path})
    {   return $handler->($conn, $req, $uri);
    }

    my $item = $tree->filename($path);

    -f $item  # filename
        and return $conn->sendFile($req, $item);

    -d _      # neither file nor directory
        or return $conn->sendStatus($req, RC_NOT_FOUND
            , "special file cannot be accessed");

    substr($item, -1) eq '/'
        or return $conn->sendRedirect($req, RC_TEMPORARY_REDIRECT, $path .'/');

    foreach my $if (@{$self->{HSV_indexfns}})
    {   return $conn->sendFile($req, $item.$if, [Location => $path.$if])
            if -f $item.$if;
    }

    $self->{HSV_dirlist}
        or return $conn->sendStatus($req, RC_FORBIDDEN, "no directory list");

    # Directory handling

    $conn->directoryList($req, $item
      , sub { my $list = shift;
              return $list if UNIVERSAL::isa($list, 'HTTP::Response');
              $self->showDirectory($conn, $req, $path, $list);
            });
}

=method showDirectory CONNECTION, REQUEST, PATH, LIST
Overrule this method with the way you would like to display an
automatically generated directory index.
=cut

sub showDirectory($$$$)
{   my ($self, $conn, $req, $dir, $list) = @_;
    my $now  = localtime;
    my @rows;
    push @rows, <<__UP if $dir ne '/';
<tr><td colspan="5">&nbsp;</td><td><a href="../">(up)</a></td></tr>
__UP

    foreach my $item (sort keys %$list)
    {   my $d = $list->{$item};
        push @rows, <<__ROW;
<tr><td>$d->{flags}</td>
    <td>$d->{user}</td>
    <td>$d->{group}</td>
    <td align="right">$d->{size_nice}</td>
    <td>$d->{mtime_nice}</td>
    <td><a href="$d->{name}">$d->{name}</a></td></tr>
__ROW
    }

    local $" = "\n";
    $conn->sendResponse($req, RC_OK, [], <<__PAGE);
<html><head><title>$dir</title></head>
<style>TD { padding: 0 10px; }</style>
<body>
<h1>Directory $dir</h1>
<table>
@rows
</table>
<p><i>Generated $now</i></p>
</body></html>
__PAGE
}

#----------------------
=section Basic daemon actions

=method rewrite URI
Returns an URI object as result, which may be the original in case of
no rewrite was needed.  See L</URI Rewrite>.
=cut

sub rewrite($) { $_[0]->{HSV_rewrite}->($_[1]) }

=method allow CLIENT, SESSION, REQUEST, URI
BE WARNED that the URI is the rewrite of the REQUEST uri, and therefore
you should use that URI.  The SESSION represents a user.

See L</Allow access>.
=cut

sub allow($$$$)
{   my ($self, $client, $session, $req, $uri) = @_;

    if($EUID==0 && substr($uri->path, 0, 2) eq '/~')
    {   notice "deamon running as root, {session} only access to {path}"
          , session => $session->id, path => '/~user';
        return 0;
    }
    1;
}

#------------------
=section Directories

=method filename URI
Translate the URI into a filename, without checking for existence.  Returns
<undef> is not possible.
=cut

sub filename($)
{   my ($self, $uri) = @_;

    my $path = $uri->path;
    my $dir = $self->directoryOf($path);
    $dir ? $dir->filename($path) : undef;
}

=method addDirectory OBJECT|OPTIONS
Either pass a M<HTTP::Server::Directory> OBJECT or the OPTIONS to create
the object.
=cut

sub addDirectory(@)
{   my $self = shift;
    my $dir  = @_==1 ? shift : HTTP::Server::Directory->new(@_);
    my $path = $dir->path || '';
    !exists $self->{HSV_dirs}{$path}
        or error __x"vhost {name} directory `{path}' defined twice"
             , name => $self->name, path => $path;
    $self->{HSV_dirs}{$path} = $dir;
}

=method directoryOf PATH
Find the best matching M<HTTP::Server::Directory> object.
=cut

sub directoryOf($)
{   my ($self, $path) = @_;
    return $self->{HSV_udirs}
        if $path =~ m!^/\~!;

    my $dirs = $self->{HSV_dirs};
    $path =~ s!/$!!;

    while(length $path)
    {   return $dirs->{$path} if $dirs->{$path};
        $path =~ s!/[^/]+$!! or return;
    }
    $dirs->{'/'} ? $dirs->{'/'} : ();
}

=section Access permissions
=cut

=chapter DETAILS

=section URI Rewrite

For each request, the M<rewrite()> method is called to see whether a
rewrite of the URI is required.  The method must return the original URI
object (the only parameter) or a new URI object.

=example rewrite URI

  package My::Virtual::Host;
  use base 'HTTP::Server::VirtualHost';

  my %lookup =
    ( '/'     => '/index-en.html'
    , '/news' => 'http://news.example.org/index.html'
    );

  sub rewrite($)
  {  my ($self, $uri) = @_;

     # with lookup table
     $uri = URI->new_abs($lookup{$uri->path}, $uri)
         if exists $lookup{$uri->path};

     # whole directory trees
     $uri = URI->new_abs('/somewhere/else'.$1, $uri)
         if $uri->path =~ m!^/some/dir(/.*|$)!;
     
     # maybe more work in the base class
     $uri->SUPER::rewrite($uri);
  }
=cut

1;
